from __future__ import annotations

import shutil
import uuid
from getpass import getuser
from pathlib import Path
from tempfile import gettempdir

import dbetto
import legendhpges as hpges
import pyg4ometry as pg4
import pygeomtools
import pytest
from legendtestdata import LegendTestData

_tmptestdir = Path(gettempdir()) / f"revertex-tests-{getuser()}-{uuid.uuid4()!s}"


@pytest.fixture(scope="session")
def tmptestdir_global():
    _tmptestdir.mkdir(exist_ok=False)
    return _tmptestdir


@pytest.fixture(scope="module")
def tmptestdir(tmptestdir_global, request):
    p = tmptestdir_global / request.module.__name__
    p.mkdir(exist_ok=True)  # note: will be cleaned up globally.
    return p


def pytest_sessionfinish(exitstatus):
    if exitstatus == 0 and Path.exists(_tmptestdir):
        shutil.rmtree(_tmptestdir)


@pytest.fixture(scope="session")
def test_data_configs():
    ldata = LegendTestData()
    ldata.checkout("8247690")
    return ldata.get_path("legend/metadata/hardware/detectors/germanium/diodes")


@pytest.fixture(scope="session", autouse=True)
def test_gdml(test_data_configs):
    test_file_dir = Path(__file__).parent

    reg = pg4.geant4.Registry()

    # get metadata and make hpges
    meta_IC = dbetto.utils.load_dict(test_data_configs + "/V99000A.yaml")
    meta_BG = dbetto.utils.load_dict(test_data_configs + "/B99000A.yaml")

    # make HPGe
    hpge_IC = hpges.make_hpge(meta_IC, registry=reg)
    hpge_BG = hpges.make_hpge(meta_BG, registry=reg)

    # create a world volume
    world_s = pg4.geant4.solid.Orb("World_s", 20, registry=reg, lunit="cm")
    world_l = pg4.geant4.LogicalVolume(world_s, "G4_Galactic", "World", registry=reg)
    reg.setWorld(world_l)

    # let's make a liquid argon balloon
    lar_s = pg4.geant4.solid.Orb("LAr_s", 15, registry=reg, lunit="cm")
    lar_l = pg4.geant4.LogicalVolume(lar_s, "G4_lAr", "LAr_l", registry=reg)
    pg4.geant4.PhysicalVolume([0, 0, 0], [0, 0, 0], lar_l, "LAr", world_l, registry=reg)

    # now place the two HPGe detectors in the argon
    hpge_IC_pv = pg4.geant4.PhysicalVolume(
        [0, 0, 0], [5, 0, -3, "cm"], hpge_IC, "V99000A", lar_l, registry=reg
    )
    hpge_BG_pv = pg4.geant4.PhysicalVolume(
        [0, 0, 0], [-5, 0, -3, "cm"], hpge_BG, "B99000A", lar_l, registry=reg
    )

    # register them as sensitive in remage
    # this also saves the metadata into the files for later use
    hpge_IC_pv.pygeom_active_detector = pygeomtools.RemageDetectorInfo(
        "germanium",
        1,
        meta_IC,
    )
    hpge_BG_pv.pygeom_active_detector = pygeomtools.RemageDetectorInfo(
        "germanium",
        2,
        meta_BG,
    )

    pygeomtools.write_pygeom(reg, f"{test_file_dir}/test_files/geom.gdml")

    return f"{test_file_dir}/test_files/geom.gdml"



from __future__ import annotations

import legendhpges
import numpy as np
from pyg4ometry import geant4

from revertex.generators.surface import (
    _sample_hpge_surface_impl,
    sample_hpge_surface,
)


def test_surface_gen(test_data_configs):
    hpge = legendhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=None)

    coords = _sample_hpge_surface_impl(100, hpge, surface_type=None, depth=None)
    assert np.shape(coords) == (100, 3)

    dist = hpge.distance_to_surface(coords)
    assert np.allclose(a=dist, b=(1e-11) * np.ones_like(dist), atol=1e-9)

    # test one surf type
    coords = _sample_hpge_surface_impl(100, hpge, surface_type="pplus", depth=None)
    assert np.allclose(a=dist, b=(1e-11) * np.ones_like(dist), atol=1e-9)

    coords = _sample_hpge_surface_impl(100, hpge, surface_type="nplus", depth=None)
    assert np.allclose(a=dist, b=(1e-11) * np.ones_like(dist), atol=1e-9)

    coords = _sample_hpge_surface_impl(100, hpge, surface_type="passive", depth=None)
    assert np.allclose(a=dist, b=(1e-11) * np.ones_like(dist), atol=1e-9)


def test_many_surface_gen(test_data_configs):
    reg = geant4.Registry()
    hpge_IC = legendhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=reg)
    hpge_BG = legendhpges.make_hpge(test_data_configs + "/B99000A.yaml", registry=reg)
    hpge_SC = legendhpges.make_hpge(test_data_configs + "/C99000A.yaml", registry=reg)

    coords = sample_hpge_surface(
        1000,
        seed=None,
        hpges={"V99000A": hpge_IC, "B99000A": hpge_BG, "C99000A": hpge_SC},
        positions={"V99000A": [0, 0, 0], "B99000A": [0, 0, 0], "C99000A": [0, 0, 0]},
    )

    assert np.shape(coords) == (1000, 3)

    # should also work for one hpge
    coords = sample_hpge_surface(1000, seed=None, hpges=hpge_IC, positions=[0, 0, 0])

    assert np.shape(coords) == (1000, 3)



from __future__ import annotations

from pathlib import Path

import legendhpges
import numpy as np
import pyg4ometry
from pyg4ometry import geant4

from revertex import utils
from revertex.utils import read_input_beta_csv


def test_read():
    path = f"{Path(__file__).parent}/test_files/beta.csv"

    energy, spec = read_input_beta_csv(path, delimiter=",")

    assert np.all(
        energy == np.array([0.0, 1.0, 2.0, 3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0])
    )
    assert np.all(
        spec == np.array([0.1, 0.2, 0.3, 0.4, 0.6, 0.5, 0.2, 0.1, 0.05, 0.07])
    )


def test_regex():
    assert utils.expand_regex(["B0", "V0"], ["B*"]) == ["B0"]


def test_hpges():
    test_file_dir = Path(__file__).parent
    gdml = f"{test_file_dir}/test_files/geom.gdml"
    reg = pyg4ometry.gdml.Reader(gdml).getRegistry()

    hpges, pos = utils.get_hpges(reg, ["B*"])

    assert isinstance(hpges["B99000A"], legendhpges.base.HPGe)
    assert isinstance(pos["B99000A"], list)

    # all
    assert len(utils.get_surface_indices(hpges["B99000A"], None)) == 8

    # just nplus
    assert len(utils.get_surface_indices(hpges["B99000A"], "nplus")) == 3

    assert utils.get_surface_weights(hpges, None)[0] == 1.0
    assert utils.get_surface_weights(hpges, "nplus")[0] == 1.0


def test_borehole(test_data_configs):
    reg = geant4.Registry()
    hpge_IC = legendhpges.make_hpge(test_data_configs + "/V99000A.yaml", registry=reg)

    borehole_vol = utils.get_borehole_volume(hpge_IC)
    assert isinstance(borehole_vol, float)

    assert utils.get_borehole_weights({"V99000A": hpge_IC})[0] == 1.0



from __future__ import annotations

import awkward as ak
import numpy as np

from revertex.generators import beta


def test_beta():
    e = np.array([0.0, 1.0, 2.0, 3.0, 5.0, 7.0, 9.0, 11.0, 13.0, 15.0])
    p = np.array([0.1, 0.2, 0.3, 0.4, 0.6, 0.5, 0.2, 0.1, 0.05, 0.07])

    assert ak.all(beta.generate_beta_spectrum(100, energies=e, phase_space=p).ekin < 15)



from __future__ import annotations

import awkward as ak
import hist
import numpy as np
from scipy import stats

from revertex import core


def test_hist_sample_one_dim():
    rng = np.random.default_rng()

    # basic case - only one bin with entries

    h = hist.Hist.new.Reg(10, 0, 10).Double().fill([0.1, 0.1])

    samples = core.sample_histogram(h, size=1000)

    # all samples in the right bin
    assert np.all((samples > 0) & (samples < 1))
    assert len(samples) == 1000

    # fill with 100 events
    n_tot = 100000
    n = 10
    h = hist.Hist.new.Reg(10, 0, n).Double().fill(rng.uniform(0, 10, size=n_tot))

    samples = core.sample_histogram(h, size=n_tot)
    h2 = hist.Hist.new.Reg(10, 0, n).Double().fill(samples)

    expected_fraction, _ = h.to_numpy()
    fraction, _ = h2.to_numpy()
    expected_fraction /= n_tot
    fraction /= n_tot

    test_stat = 2 * np.sum(
        n_tot * expected_fraction
        - (fraction * n_tot) * (1 - np.log(fraction / expected_fraction))
    )
    p = stats.chi2.sf(test_stat, n - 1)
    sigma = stats.norm.ppf(1 - p)

    assert sigma < 5


def test_hist_sample_two_dim():
    rng = np.random.default_rng()

    # basic case - only one bin with entries

    h = (
        hist.Hist.new.Reg(10, 0, 10)
        .Reg(10, 0, 10)
        .Double()
        .fill([0.1, 0.1], [0.1, 0.1])
    )

    samples_x, samples_y = core.sample_histogram(h, size=1000)

    # all samples in the right bin
    assert np.all((samples_x > 0) & (samples_x < 1))
    assert np.all((samples_y > 0) & (samples_y < 1))

    assert len(samples_x) == 1000
    assert len(samples_y) == 1000

    # fill with 100 events
    n_tot = 1000000
    n = 10
    h = (
        hist.Hist.new.Reg(10, 0, n)
        .Reg(10, 0, n)
        .Double()
        .fill(rng.uniform(0, 10, size=n_tot), rng.uniform(0, 10, size=n_tot))
    )
    sample_x, sample_y = core.sample_histogram(h, size=n_tot)
    h2 = hist.Hist.new.Reg(10, 0, n).Reg(10, 0, n).Double().fill(sample_x, sample_y)
    expected_fraction, _, _ = h.to_numpy()
    fraction, _, _ = h2.to_numpy()

    # flatten
    fraction = fraction.flatten()
    expected_fraction = expected_fraction.flatten()

    # normalise
    expected_fraction /= n_tot
    fraction /= n_tot

    test_stat = 2 * np.sum(
        n_tot * expected_fraction
        - (fraction * n_tot) * (1 - np.log(fraction / expected_fraction))
    )
    p = stats.chi2.sf(test_stat, n * n - 1)
    sigma = stats.norm.ppf(1 - p)

    assert sigma < 5


def test_convert():
    arr = ak.Array({"xloc": [1, 2, 3], "yloc": [1, 2, 3], "zloc": [1, 2, 3]})

    for f in arr.fields:
        assert ak.all(core.convert_output(arr, mode="pos").view_as("ak")[f] == arr[f])

    arr = ak.Array(
        {
            "px": [1, 2, 3],
            "py": [1, 2, 3],
            "pz": [1, 2, 3],
            "ekin": [1, 2, 3],
            "g4_pid": [11, 11, 11],
        }
    )

    for f in arr.fields:
        assert ak.all(core.convert_output(arr, mode="kin").view_as("ak")[f] == arr[f])


def test_sample_proportional_radius():
    samples = core.sample_proportional_radius(
        np.zeros(10000), np.ones(10000), size=10000
    )
    assert len(samples) == 10000


def test_sample_cylinder():
    samples = core.sample_cylinder((0, 10), (-1, 11), 100, None)
    assert samples.shape == (100, 3)



from __future__ import annotations

import argparse
import logging

import pyg4ometry

from revertex import core, utils
from revertex.generators import beta, borehole, shell, surface
from revertex.utils import setup_log

log = logging.getLogger(__name__)


def cli(args=None) -> None:
    parser = argparse.ArgumentParser(
        prog="revertex",
        description="%(prog)s command line interface",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=1,
        help="""Increase the program verbosity""",
    )
    parser.add_argument(
        "--seed",
        "-s",
        default=None,
        type=int,
        help="Seed for rng",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    # beta spectra
    beta_parser = subparsers.add_parser(
        "beta-kin", help="Generate beta kinematics from a csv file."
    )

    beta_parser.add_argument(
        "--input-file",
        "-i",
        required=True,
        type=str,
        help="Path to the file with the spectrum to sample",
    )
    beta_parser.add_argument(
        "--out-file",
        "-o",
        required=True,
        type=str,
        help="Path to the output file. ",
    )
    beta_parser.add_argument(
        "--n-events",
        "-n",
        required=True,
        type=int,
        help="Number of events to generate",
    )
    beta_parser.add_argument(
        "--eunit",
        "-e",
        required=True,
        type=str,
        help="Unit for the energy",
    )

    hpge_surface_parser = subparsers.add_parser(
        "hpge-surf-pos", help="Generate samples from the surface of the HPGes"
    )

    hpge_surface_parser.add_argument(
        "--gdml",
        "-g",
        required=True,
        type=str,
        help="Path to the GDML file of the geometry",
    )
    hpge_surface_parser.add_argument(
        "--surface-type",
        "-t",
        required=True,
        type=str,
        help="Type of surface",
    )
    hpge_surface_parser.add_argument(
        "--detectors",
        "-d",
        required=True,
        nargs="+",
        type=str,
        help="Name of a detector, list of detectors or regex's.",
    )
    hpge_surface_parser.add_argument(
        "--out-file",
        "-o",
        required=True,
        type=str,
        help="Path to the output file. ",
    )
    hpge_surface_parser.add_argument(
        "--n-events",
        "-n",
        required=True,
        type=int,
        help="Number of events to generate",
    )

    hpge_shell_parser = subparsers.add_parser(
        "hpge-shell-pos", help="Generate samples from the shell of the HPGes"
    )

    hpge_shell_parser.add_argument(
        "--gdml",
        "-g",
        required=True,
        type=str,
        help="Path to the GDML file of the geometry",
    )
    hpge_shell_parser.add_argument(
        "--surface-type",
        "-t",
        required=True,
        type=str,
        help="Type of surface",
    )
    hpge_shell_parser.add_argument(
        "--detectors",
        "-d",
        required=True,
        nargs="+",
        type=str,
        help="Name of a detector, list of detectors or regex's.",
    )
    hpge_shell_parser.add_argument(
        "--out-file",
        "-o",
        required=True,
        type=str,
        help="Path to the output file. ",
    )
    hpge_shell_parser.add_argument(
        "--n-events",
        "-n",
        required=True,
        type=int,
        help="Number of events to generate",
    )
    hpge_shell_parser.add_argument(
        "--radius",
        "-r",
        required=True,
        type=float,
        help="Radius of the shell to generate in",
    )

    hpge_borehole_parser = subparsers.add_parser(
        "hpge-borehole-pos", help="Generate samples from the borehole of the HPGes"
    )

    hpge_borehole_parser.add_argument(
        "--gdml",
        "-g",
        required=True,
        type=str,
        help="Path to the GDML file of the geometry",
    )

    hpge_borehole_parser.add_argument(
        "--detectors",
        "-d",
        required=True,
        nargs="+",
        type=str,
        help="Name of a detector, list of detectors or regex's.",
    )
    hpge_borehole_parser.add_argument(
        "--out-file",
        "-o",
        required=True,
        type=str,
        help="Path to the output file. ",
    )
    hpge_borehole_parser.add_argument(
        "--n-events",
        "-n",
        required=True,
        type=int,
        help="Number of events to generate",
    )

    args = parser.parse_args(args)

    log_level = (None, logging.INFO, logging.DEBUG)[min(args.verbose, 2)]
    setup_log(log_level)

    if args.command == "beta-kin":
        msg = f"Generating beta kinematics from {args.input_file} to {args.out_file} and seed {args.seed}"
        log.info(msg)

        beta.save_beta_spectrum(
            n_gen=args.n_events,
            out_file=args.out_file,
            in_file=args.input_file,
            seed=args.seed,
            eunit=args.eunit,
        )

    elif args.command == "hpge-surf-pos":
        msg = "Generating points on the HPGes for \n"
        msg += f"gdml:      {args.gdml} \n"
        msg += f"output:    {args.out_file} \n"
        msg += f"seed:      {args.seed} \n"
        msg += f"detectors: {args.detectors} ({args.surface_type})"
        log.info(msg)

        # read the registry
        reg = pyg4ometry.gdml.Reader(args.gdml).getRegistry()

        hpges, pos = utils.get_hpges(reg, args.detectors)

        core.write_remage_vtx(
            args.n_events,
            args.out_file,
            args.seed,
            surface.sample_hpge_surface,
            hpges=hpges,
            positions=pos,
            surface_type=args.surface_type,
        )
    elif args.command == "hpge-shell-pos":
        msg = "Generating points on the HPGes shells for \n"
        msg += f"gdml:      {args.gdml} \n"
        msg += f"output:    {args.out_file} \n"
        msg += f"seed:      {args.seed} \n"
        msg += f"detectors: {args.detectors} ({args.surface_type})"
        msg += f"radius : {args.radius}"
        log.info(msg)

        # read the registry
        reg = pyg4ometry.gdml.Reader(args.gdml).getRegistry()

        hpges, pos = utils.get_hpges(reg, args.detectors)

        core.write_remage_vtx(
            args.n_events,
            args.out_file,
            args.seed,
            shell.sample_hpge_shell,
            hpges=hpges,
            positions=pos,
            distance=args.radius,
            surface_type=args.surface_type,
        )
    elif args.command == "hpge-borehole-pos":
        msg = "Generating points on the HPGes boreholes for \n"
        msg += f"gdml:      {args.gdml} \n"
        msg += f"output:    {args.out_file} \n"
        msg += f"seed:      {args.seed} \n"
        msg += f"detectors: {args.detectors}"
        log.info(msg)

        # read the registry
        reg = pyg4ometry.gdml.Reader(args.gdml).getRegistry()

        hpges, pos = utils.get_hpges(reg, args.detectors)

        core.write_remage_vtx(
            args.n_events,
            args.out_file,
            args.seed,
            borehole.sample_hpge_borehole,
            hpges=hpges,
            positions=pos,
        )



from __future__ import annotations

from pathlib import Path

from lgdo import lh5

from revertex.cli import cli


def test_cli(tmptestdir, test_gdml):
    test_file_dir = Path(__file__).parent

    # test cli for betas
    cli(
        [
            "beta-kin",
            "-i",
            f"{test_file_dir}/test_files/beta.csv",
            "-o",
            f"{tmptestdir}/test_beta.lh5",
            "-n",
            "2000",
            "-e",
            "keV",
        ]
    )

    kin = lh5.read("vtx/kin", f"{tmptestdir}/test_beta.lh5").view_as("ak")
    assert set(kin.fields) == {"px", "py", "pz", "ekin", "g4_pid"}
    assert len(kin) == 2000

    cli(
        [
            "hpge-surf-pos",
            "-g",
            test_gdml,
            "-t",
            "nplus",
            "-d",
            "B*",
            "-o",
            f"{tmptestdir}/test_surf.lh5",
            "-n",
            "1000",
        ]
    )

    pos = lh5.read("vtx/pos", f"{tmptestdir}/test_surf.lh5").view_as("ak")
    assert set(pos.fields) == {"xloc", "yloc", "zloc"}

    assert len(pos) == 1000

    cli(
        [
            "hpge-shell-pos",
            "-g",
            test_gdml,
            "-t",
            "nplus",
            "-d",
            "B*",
            "-r",
            "2",
            "-o",
            f"{tmptestdir}/test_shell.lh5",
            "-n",
            "1000",
        ]
    )

    pos = lh5.read("vtx/pos", f"{tmptestdir}/test_shell.lh5").view_as("ak")
    assert set(pos.fields) == {"xloc", "yloc", "zloc"}

    assert len(pos) == 1000

    cli(
        [
            "hpge-borehole-pos",
            "-g",
            test_gdml,
            "-d",
            "V*",
            "-o",
            f"{tmptestdir}/test_bh.lh5",
            "-n",
            "1000",
        ]
    )

    pos = lh5.read("vtx/pos", f"{tmptestdir}/test_bh.lh5").view_as("ak")
    assert set(pos.fields) == {"xloc", "yloc", "zloc"}

    assert len(pos) == 1000
